//
//  SubmitPanelController.m
//  Perforce
//
//  Created by Adam Czubernat on 10.07.2013.
//  Copyright (c) 2013 Perforce Software, Inc. All rights reserved.
//

#import "SubmitPanelController.h"
#import "P4Workspace.h"
#import "P4ChangelistItem.h"
#import "ErrorPanelController.h"

static NSString *lastMessage;

@interface SubmitPanelController () <NSTableViewDataSource, NSTableViewDelegate, NSTextViewDelegate, P4ItemDelegate> {
	
	P4Item *root;
	NSMutableArray *items;
	NSMutableSet *selectedItems;
    NSMutableArray *selectedPaths;
	
	NSUInteger totalSize;
	BOOL indeterminate;
	
	__weak IBOutlet NSView *submitView;
	__weak IBOutlet NSView *indeterminateView;
	__weak IBOutlet NSView *progressView;
		
	__weak IBOutlet NSTableView *tableView;
	__weak IBOutlet NSProgressIndicator *tableIndicator;
	__unsafe_unretained IBOutlet NSTextView *descriptionTextView;
	__weak IBOutlet NSTextField *fileCountLabel;
	__weak IBOutlet NSButton *submitButton;
		
	__weak IBOutlet NSProgressIndicator *indeterminateIndicator;
	__weak IBOutlet NSTextField *indeterminateLabel;
	__weak IBOutlet NSImageView *successImageView;
	
	__weak IBOutlet NSProgressIndicator *progressIndicator;
	__weak IBOutlet NSTextField *progressLabel;
	__weak IBOutlet NSImageView *progressImageView;
}

- (void)reload;

- (void)updateCount;
- (void)validateSubmit;
- (void)setItem:(P4ChangelistItem *)item selected:(BOOL)selected;

- (void)submit;
- (void)showProgress;
- (void)updateProgress:(P4Operation *)operation;
- (void)showErrors:(P4Operation *)operation;

- (IBAction)checkmarkPressed:(id)sender;
- (IBAction)checkmarkHeaderPressed:(id)sender;
- (IBAction)cancelPressed:(id)sender;
- (IBAction)submitPressed:(id)sender;
@end

@implementation SubmitPanelController

- (void)windowDidLoad {
	[super windowDidLoad];

	[submitButton setEnabled:NO];
	fileCountLabel.stringValue = @"";
	descriptionTextView.string = lastMessage ? : @"";
	[tableView setSortDescriptors:@[ [NSSortDescriptor
									  sortDescriptorWithKey:
									  @"remotePath.stringByDeletingLastPathComponent"
									  ascending:YES] ]];
	
	[self setPresentedView:submitView];
	[self reload];
	
	[P4ChangelistItem addObserver:self];
}

- (void)dealloc {
	[P4ChangelistItem removeObserver:self];
}

#pragma mark - Private

- (void)reload {
	
	root = nil;
	items = nil;
	selectedItems = nil;
	
	[tableView reloadData];
	[tableIndicator startAnimation:nil];
	
	root = [[P4ChangelistItem alloc] initWithParent:nil];
	[root children]; // Load children
}

- (void)updateCount {
	fileCountLabel.stringValue = [NSString stringWithFormat:@"Selected %ld / %ld",
								  selectedItems.count, items.count];
}

- (void)validateSubmit {
	BOOL enable = descriptionTextView.string.length && selectedItems.count;
	[submitButton setEnabled:enable];
}

- (void)setItem:(P4ChangelistItem *)item selected:(BOOL)selected {
	if (selected) {
		[selectedPaths addObject:item.remotePath.lowercaseString];
		[selectedItems addObject:item];
	} else {
		[selectedPaths removeObject:item.remotePath.lowercaseString];
		[selectedItems removeObject:item];
	}
	NSInteger row = [items indexOfObjectIdenticalTo:item];
	[tableView reloadDataForRowsFromIndex:row toIndex:row];
}

- (void)submit {
	
	NSString *description = descriptionTextView.string;
	
	indeterminate = YES;
	[indeterminateIndicator setHidden:NO];
	[indeterminateIndicator startAnimation:nil];
	[self setPresentedView:indeterminateView animated:YES];
	
	totalSize = 0;
	NSMutableArray *files = [NSMutableArray array];
	NSFileManager *filemanager = [NSFileManager defaultManager];
	for (P4Item *item in selectedItems) {
		if (item.remotePath)
			[files addObject:item.remotePath];
		if ([item.status isEqualToString:@"add"] ||
			[item.status isEqualToString:@"edit"])
			totalSize += [[filemanager attributesOfItemAtPath:item.localPath
														error:NULL] fileSize];
	}
	
	[[P4Workspace sharedInstance]
	 submitFiles:files
	 message:description
	 receive:^(P4Operation *operation) {
		if (indeterminate)
			 [self showProgress]; // Switch to progress
		 [self updateProgress:operation];
	 }
	 response:^(P4Operation *operation, NSArray *response) {
		 
		 [progressImageView.layer removeAllAnimations];
		 [indeterminateIndicator stopAnimation:nil];
		 [indeterminateIndicator setHidden:YES];
		 
		 if (operation.errors) {
			 
			 [self showErrors:operation];
			 
		 } else {
			 
			 progressLabel.stringValue = @"Completed";
			 indeterminateLabel.stringValue = @"Success!";
			 [successImageView setHidden:NO];

			 if (!indeterminate)
				 [self setPresentedView:indeterminateView animated:YES];
			 
			 lastMessage = nil;
			 PSLog(@"%@", response);
			 
			 [self dismissAfter:2.0f];
		 }
	 }];
}

- (void)showProgress {
	indeterminate = NO;
	[indeterminateIndicator stopAnimation:nil];
	[self setPresentedView:progressView animated:YES];
	
	// Animate image
	progressImageView.wantsLayer = YES;
	CAKeyframeAnimation *anim = [CAKeyframeAnimation animationWithKeyPath:@"transform.translation.y"];
	CAMediaTimingFunction *timing = [CAMediaTimingFunction functionWithName:kCAMediaTimingFunctionEaseInEaseOut];
	anim.duration = 2.0f;
	anim.repeatCount = HUGE_VALF;
    anim.values	  =	@[ @(0.0f), @(3.0f), @(0.0) ];
    anim.keyTimes = @[ @(0.0f), @(0.4f), @(0.8) ];
    anim.timingFunctions = @[ timing, timing ];
	anim.removedOnCompletion = NO;
	[progressImageView.layer addAnimation:anim forKey:@"translate"];
	progressImageView.layer.masksToBounds = NO;
}

- (void)updateProgress:(P4Operation *)operation {
	
	CGFloat progress = fmin(operation.completed / (CGFloat)totalSize, 1.0f);
	progressIndicator.doubleValue = progress * 100.0f;
	NSTimeInterval time = PSTimeInterval(operation.timestamp);
	NSTimeInterval eta = fmax(time / progress - time, 0.0f);
	
	progressLabel.stringValue = [NSString stringWithFormat:
								 @"%.1f%% (%.2f of %.2f MB) – %@ remaining",
								 progress * 100.0f,
								 operation.completed/1024/1024.0f,
								 totalSize/1024/1024.0f,
								 [NSString stringWithTimeInterval:eta]];
}

- (void)showErrors:(P4Operation *)operation {

	[self dismiss];
	
	[operation ignoreErrorsWithCode:P4ErrorSessionExpired]; // Don't show alert for session expiration
	[operation ignoreErrorsWithCode:P4ErrorSubmitAborted];
	[operation ignoreErrorsWithCode:P4ErrorSubmitUnresolved];
	NSArray *resolveErrors = [operation errorsWithCode:P4ErrorResolve];
	[operation ignoreErrors:resolveErrors];
	
	NSMutableArray *details = [NSMutableArray array];
	if (resolveErrors) {
		[details addObject:@"Unresolved"];
		[details addObjectsFromArray:[resolveErrors valueForKeyPath:@"localizedDescription"]];
		[details addObject:@"\n"];
	}
	if (operation.errors) {
		[details addObject:@"Submit encountered errors"];
		[details addObjectsFromArray:[operation.errors valueForKeyPath:@"localizedDescription"]];
	}
	
	if (!details.count)
		return; // Nothing to display

	ErrorPanelController *alert = [[ErrorPanelController alloc] init];
	[alert setTitle:@"\nSubmit failed"];
	[alert setMessage:resolveErrors ?
	 @"Submit failed because some of selected files need to be resolved" :
	 @"Submit failed with errors"];
	[alert setDetails:[details componentsJoinedByString:@"\n"]];
	[alert setDetailsTitle:[NSString stringWithFormat:@"Error details (%ld)",
							resolveErrors.count + operation.errors.count]];
	[alert setCompletionBlock:^{
		[self setPresentedView:submitView animated:NO];
		[self reload];
		[self presentWithMainWindow];
	}];
	[alert presentWithMainWindow];
}

#pragma mark - Public

- (void)selectPaths:(NSArray *)paths {
	P4Workspace *workspace = [P4Workspace sharedInstance];
	
	// Get remote paths
	selectedPaths = [NSMutableArray array];
	for (__strong NSString *path in paths) {
		if (![path hasPrefix:@"//"])
			path = [workspace mappingForPath:path];
		if (path)
			[selectedPaths addObject:path.lowercaseString];
	}
	
	if (items)
		[self reload];
}

#pragma mark - P4Item delegate

- (void)itemDidLoad:(id)item {
	if (item != root)
		return;

	NSInteger visibleRow = 0;
	selectedItems = [NSMutableSet set];
	items = [root.children mutableCopy];
	[items sortUsingDescriptors:[tableView sortDescriptors]];
	
    if (selectedPaths.count > 0) {
		// Select items from paths
		NSArray *selected = [items filteredArrayUsingPredicate:
							 [NSPredicate predicateWithFormat:
							  @"remotePath.lowercaseString IN %@ OR "
							  "metadata.movedFile.lowercaseString IN %@", // Check moved files
							  selectedPaths, selectedPaths]];
		[selectedItems addObjectsFromArray:selected];
		
		// Get first selected row
		if (selected.count)
			visibleRow = [items indexOfObjectIdenticalTo:[selected objectAtIndex:0]];
		
    } else {
		// Select all
		selectedItems = [NSMutableSet setWithArray:items];
	}
	
	[self updateCount];
	[self validateSubmit];
	
	[tableView reloadData];
	[tableIndicator stopAnimation:nil];
	[tableView scrollRowToVisible:visibleRow];
}

- (void)itemDidInvalidate:(id)item {
	
}

- (void)item:(id)item didFailWithError:(NSError *)error {
	
}

#pragma mark - TableView data source

- (NSInteger)numberOfRowsInTableView:(NSTableView *)tableView {
	return items.count;
}

- (NSView *)tableView:(NSTableView *)table viewForTableColumn:(NSTableColumn *)tableColumn row:(NSInteger)row {

	NSString *identifier = tableColumn.identifier;
	P4ChangelistItem *item = [items objectAtIndex:row];
	NSTableCellView *cell = [tableView makeViewWithIdentifier:identifier owner:self];
	
	if ([identifier isEqualToString:@"checkmark"]) {
		NSButton *checkmark = [cell viewWithTag:1];
		NSAssert(checkmark, @"Submit panel cell should have checkmark with tag=1");
		checkmark.state = [selectedItems containsObject:item];
	} else if ([identifier isEqualToString:@"filename"]) {
		cell.textField.stringValue = [item name];
		cell.imageView.image = item.icon;
	} else if ([identifier isEqualToString:@"path"]) {
		NSString *path = [item remotePath];
		cell.textField.stringValue = [path stringByDeletingLastPathComponent];
	} else if ([identifier isEqualToString:@"action"]) {
		NSString *action = item.status;
		cell.textField.stringValue = action;
		if ([item.metadata objectForKey:@"unresolved"])
			cell.textField.stringValue = [action stringByAppendingString:@"/unresolved"];
	}
	
	return cell;
}

- (void)tableView:(NSTableView *)aTableView sortDescriptorsDidChange:(NSArray *)oldDescriptors {
	[items sortUsingDescriptors:[tableView sortDescriptors]];
	[tableView reloadData];
}

#pragma mark - TextView delegate

- (void)textDidChange:(NSNotification *)notification {
	lastMessage = descriptionTextView.string;
	[self validateSubmit];
}

#pragma mark - Actions

- (IBAction)checkmarkPressed:(NSButton *)sender {
	
	BOOL selected = sender.state == NSOnState;
	NSInteger row = [tableView rowForView:sender];
	P4ChangelistItem *child = [items objectAtIndex:row];

	// Add or remove from selected list
	[self setItem:child selected:selected];
		
	// Select moved counterpart
	if ([child.status hasPrefix:@"move/"]) {
		P4ChangelistItem *moved = [[items filteredArrayUsingPredicate:
									[NSPredicate predicateWithFormat:
									 @"metadata.movedFile == %@",
									 child.remotePath]] lastObject];
		[self setItem:moved selected:selected];
	}
	
	[self updateCount];
	[self validateSubmit];
}

- (IBAction)checkmarkHeaderPressed:(NSButton *)sender {
	if (sender.state)
		[selectedItems addObjectsFromArray:items];
	else
		[selectedItems removeAllObjects];

	[tableView reloadData];
	[self updateCount];
	[self validateSubmit];
}

- (IBAction)cancelPressed:(id)sender {
	[self dismiss];
}

- (IBAction)submitPressed:(id)sender {
	[self submit];
}

@end
